#!/usr/bin/python

# ---> Imports
import base64
import curses
import email
import errno
import os
import sys
import pickle
import re
import socket
import logging

from apiclient import errors
from pmail.common import (mkService, Labels, MessageInfo, Attachments,
                    listMessagesMatchingQuery, logger,
                    UserInfo, AddressBook, config)
from itertools import cycle
from pmail.sendmail import (sendMessage, mkSubject, mkTo, createMessage)
from subprocess import run, PIPE, Popen
from threading import Thread, Event, Lock
from uuid import uuid4
from time import sleep, process_time
from queue import Queue
# <---

# ---> Curses functions


def initColors():
  curses.start_color()
  curses.use_default_colors()
  # Color pair for main list.
  curses.init_pair(1, config.fg, config.bg)

  # Color pair for selected messages.
  curses.init_pair(2, config.selFg, config.selBg)

  # Color pair for hightlighted messages.
  curses.init_pair(3, config.hiFg, config.hiBg)

  # Color pair for hightlighted and selected messages.
  curses.init_pair(4, config.selFg, config.hiBg)

  # Color pair for status bar.
  curses.init_pair(5, config.stFg, config.stBg)

# ---> Getting input and displaying messages


def getInput(stdscr, prompt, height, width):
  '''
  Get input from the user.

  Args:
    stdscr: The curses window object.
    prompt: Prompt to show the user.
    height: Height of stdscr.
    width: Width of stdscr.

  Returns:
    A string of charecters input by the user, or none of the user aborts
    by pressing ESCAPE.
  '''
  address = ""
  k, p = 0, len(prompt)
  stdscr.addstr(height - 1, 0, ' ' * (width - 1))
  stdscr.refresh()
  curses.curs_set(1)
  while k != '\n':
    l = len(address)
    stdscr.addstr(height - 1, 0, prompt + address)
    stdscr.addstr(height - 1, l + p, ' ' * (width - 1 - l - p))
    stdscr.move(height - 1, p + l)
    stdscr.refresh()
    k = stdscr.getkey()
    if (k == 'KEY_BACKSPACE'):
      try:
        address = address[:l-1]
      except:
        pass
    elif (len(k) == 1) and (k != '\t'):
      if ord(k) == 27:
        return None
      else:
        address += k
  return address[:-1]


def putMessage(stdscr, height, width, message):
  '''
  Display a message at the bottom of the screen.

  Args:
    stdscr: The curses window object.
    height: Height of stdscr.
    width: Width of stdscr.
    message: The message to display

  Returns: None.
  '''
  line = (":: " + message + ' ' * width)[:width - 1]
  stdscr.addstr(height - 1, 0, line)

# <---

# ---> Statusbar class


'''
class StatusBarInfo():
  Class to store various data needed to render the status bar and a
  method to render it.

  Args:
    account: Currently selected account.
    includedLabels: Labels included in the filter in __getHeaders.
    excludedLabels: Labels excluded by the filter in __getHeaders.
    selectedHeader: The hightlighted message.
    searchTerms: The terms searched for, if any.
    numOfMessages: The number of messages in the query returned by __getHeaders.
    totalMessages: The total num of messages in account. (not used atm)
    currentMessageIndex: The index of the highlighted message.
    showLabels: A list of labels attached to the highlighted message.

  def __init__(self, currentAccount, includedLabels,
               excludedLabels, selectedHeader, searchTerms,
               numOfMessages, totalMessages, currentMessageIndex,
               showLabels):
    self.currentAccount = currentAccount
    self.includedLabels = includedLabels
    self.excludedLabels = excludedLabels
    self.selectedHeader = selectedHeader
    self.searchTerms = searchTerms
    self.numOfMessages = numOfMessages
    self.totalMessages = totalMessages
    self.currentMessageIndex = currentMessageIndex
    self.showLabels = showLabels

  def mkStatusBar(self, stdscr, height, width):
    Draw the status bar on the screen.

    Args:
      stdscr: The curses window object.
      height: Height of stdscr.
      width: Width of stdscr.

    Returns: None
    # user = ' \uf007 '
    # seperator = '\uf054'
    user = ' {} '.format(chr(config.user))
    seperator = chr(config.seperator)
    lUser = len(user)

    account = ' {} {} {} '.format(
        seperator,
        self.currentAccount,
        seperator)
    acLen = len(account) + lUser

    if self.searchTerms:
      st = '?[{}] {} '.format(
          self.searchTerms,
          seperator)
    else:
      st = ''
    stLen = len(st)

    mb = self.includedLabels[0]
    mbLen = len(mb)

    numOfMessages = ' {} [{}/{} messages]'.format(
        seperator,
        self.currentMessageIndex,
        self.numOfMessages)
    n = len(numOfMessages)

    if self.showLabels == True:
      labels = str(self.selectedHeader.showLabels())
    else:
      labels = ''
    labelsLen = len(labels)

    whitespace = ' {}'.format(
        ' ' * (width - mbLen - acLen - stLen - n - 1))

    stdscr.attron(curses.color_pair(5))
    stdscr.attron(curses.A_BOLD)
    stdscr.addstr(height - 2, 0, user)
    stdscr.attroff(curses.A_BOLD)

    stdscr.addstr(height - 2, lUser, account)

    stdscr.addstr(height - 2, acLen, st)

    stdscr.attron(curses.A_BOLD)
    stdscr.addstr(height - 2, acLen + stLen, mb)
    stdscr.attroff(curses.A_BOLD)

    stdscr.addstr(height - 2, acLen + stLen + mbLen, numOfMessages)

    stdscr.addstr(height - 2, acLen + stLen + mbLen + n, labels)

    stdscr.addstr(height - 2, acLen + stLen +
                  mbLen + n + labelsLen, whitespace)
    stdscr.attroff(curses.color_pair(5))

    if self.numOfMessages > 0:
      snippet = ' On {} <{}> wrote: {} {}'.format(
          self.selectedHeader.timeForReply(),
          self.selectedHeader.parseSender()[0],
          self.selectedHeader.snippet,
          ' ' * width
      )
      try:
        stdscr.addstr(height - 1, 0, snippet[:width - 1])
      except:
        stdscr.addstr(height - 1, 0, ' ' * (width - 1))
'''

# <---


''' probbaly won't include this
def showSnippet(stdscr, height, width, snippet):
  k = 0
  while k != ord('q'):
    height, width = stdscr.getmaxyx()
    w = min(80 + width % 2, width - 2)
    indent = (width - w)//2
    lines = wrap(snippet, width=(w - 2))
    top = (height - len(lines))//2
    for i, l in enumerate(lines):
      stdscr.attron(curses.color_pair(25))
      stdscr.addstr(top+i, indent, ' ' + l + ' ')
      ln = len(l)
      stdscr.addstr(top+i, indent + ln + 2, ' '*(w - ln - 2))
      stdscr.attroff(curses.color_pair(25))
    stdscr.refresh()
    k = stdscr.getch()


def search(results,excludedLabels,includedLabels):
    return _getHeaders(results,excludedLabels,includedLabels)
'''

# ---> Confirmation Screen


# def drawConfirmationScreen(stdscr, account, state):
def drawConfirmationScreen(stdscr, account, sendAction):
  '''
  Loop to draw the confirmation screen that appears before email
  is finally sent.

  Args:
    stdscr: The curses window object.
    account: Currently selected account.
    state: The state of the program.

  Returns:
    state: The state of the program.
  '''
  k = 0
  curses.curs_set(0)

  attachments = []
  cc = ''
  bcc = ''
  stdscr.clear()
  stdscr.refresh()

  while 1:
    height, width = stdscr.getmaxyx()

    l = len(attachments)
    options = {
        'y': 'send',
        'q': 'abort',
        'a': 'add attachment',
        'c': 'add recipients',
        's': 'edit subject',
        't': 'edit recipient'}
    c = 1  # counter for number of lines of infomation appearing.

    if k == curses.KEY_RESIZE:
      height, width = stdscr.getmaxyx()
      stdscr.refresh()

    elif k == ord('q'):
      # Don't send.
      sendAction.send = 'DO_NOT_SEND'
      return sendAction

    elif k == ord('y'):
      # Send.
      sendAction.send = 'SEND'
      sendAction.attachments = attachments if l > 0 else None
      sendAction.cc = cc if len(cc) > 0 else None
      sendAction.bcc = bcc if len(bcc) > 0 else None
      return sendAction

    elif k == ord('a'):
      # Add attachments.
      a = chooseAttachment(os.environ.get('HOME'))
      if a:
        attachments.append(a)
      stdscr.clear()
      stdscr.refresh()
      k = 0
      continue

    elif k == ord('c'):
      # Add recipients.
      # Either To, Bcc, or Cc
      # redraw options menu
      options = {
          't': 'To',
          'c': 'Cc',
          'b': 'Bcc'}
      stdscr.attron(curses.color_pair(1))
      for i in range(c + l + 1, height - 2):
        stdscr.addstr(i, 0, ' ' * width)
      for i, key in enumerate(options.keys()):
        stdscr.addstr(c + i + l + 1, 3, '- {}: {}'.format(key,
                                                          options[key])[:width - 2])
      stdscr.attroff(curses.color_pair(1))
      k = stdscr.getch()
      to = chooseAddress(account)
      if to:
        if k == ord('t'):
          sendAction.sendTo += (', ' + to)
        elif k == ord('c'):
          lcc = len(cc)
          cc += (', ' + to) if lcc > 0 else to
        elif k == ord('b'):
          lbcc = len(bcc)
          bcc += (', ' + to) if lbcc > 0 else to
      stdscr.clear()
      stdscr.refresh()
      k = 0
      continue

    elif k == ord('s'):
      # Edit subject.
      subject = getInput(stdscr, 'Enter new subject: ', height, width)
      stdscr.addstr(height - 1, 0, ' ' * (width - 1))
      curses.curs_set(0)
      sendAction.subject = subject
      k = 0
      continue

    elif k == ord('t'):
      # Edit recipient.
      to = chooseAddress(account)
      if to:
        sendAction.sendTo = to
      stdscr.clear()
      stdscr.refresh()
      k = 0
      continue

    stdscr.attron(curses.color_pair(1))
    stdscr.attron(curses.A_BOLD)
    stdscr.addstr(c, 1, ("You are about to send an email." +
                         " Please verify the details and press 'y' to send.")[:width - 2])
    stdscr.attroff(curses.A_BOLD)
    c += 1
    stdscr.addstr(c, 0, ' ' * (width))
    c += 1
    # Handle case when there are many to's..
    stdscr.addstr(c, 1, ("To: " + sendAction.sendTo)[:width - 2])
    c += 1
    if cc != '':
      stdscr.addstr(c, 1, ("Cc: " + cc)[:width - 2])
      c += 1
    if bcc != '':
      stdscr.addstr(c, 1, ("Bcc: " + bcc)[:width - 2])
      c += 1
    stdscr.addstr(c, 1, ("From: " + account)[:width - 2])
    c += 1
    stdscr.addstr(
        c, 1, ("Subject: " + sendAction.subject + ' ' * width)[:width - 2])
    c += 1
    stdscr.addstr(c, 1, ("Attachments: ")[:width - 2])
    c += 1

    # listAttachments(attachments)
    for i, a in enumerate(attachments):
      stdscr.addstr(c+i, 3, ('- ' + a)[:width - 4])

    stdscr.addstr(c + l, 1, ("Options:")[:width - 2])

    for i, key in enumerate(options.keys()):
      stdscr.addstr(c + i + l + 1, 3, '- {}: {}'.format(key,
                                                        options[key])[:width - 4])

    stdscr.attroff(curses.color_pair(1))
    stdscr.refresh()
    k = stdscr.getch()

# <---

# ---> Attachments


def drawAttachments(stdscr, account, state, attachments):
  '''
  Draw a list of attachments.
  Currently not handerlin the case when len(attachments) > height.

  Args:
    stdscr: The curses window object.
    account: The currently selected account.
    state: The state of the program.
    attachments: The list of attachments from the highlighted message.

  Returns:
    state: The state of the program.
  '''
  k = 0
  cursor_y = 0
  # Clear and refresh the screen for a blank canvas
  # Start colors in curses
  curses.curs_set(0)
  initColors()
  stdscr.clear()
  stdscr.refresh()

  numOfAttachments = len(attachments)
  selectedAttachment = attachments[0]
  # Loop where k is the last character pressed
  while 1:
    height, width = stdscr.getmaxyx()

    if k == curses.KEY_RESIZE:
      height, width = stdscr.getmaxyx()
      stdscr.addstr(height - 1, 0, ' ' * (width - 1))

    elif k in [curses.KEY_DOWN, ord('j')]:
      cursor_y = cursor_y + 1

    elif k in [curses.KEY_UP, ord('k')]:
      cursor_y = cursor_y - 1

    elif k == ord('s'):
      # Try to save selectedAttachment.
      saveAttachment(mkService(account), selectedAttachment)
      state.action.saved = 'SAVED_ATTACHMENT'
      return state

    elif k == ord('q'):
      state.action.saved = 'NOT_SAVED'
      return state

    cursor_y = max(0, cursor_y)
    cursor_y = min(height-3, cursor_y, numOfAttachments - 1)
    # Update selected header.
    selectedAttachment = attachments[cursor_y]

    for i, a in enumerate(attachments[:height - 2]):
      display = a.display()
      l1 = len(display)
      stdscr.attron(curses.color_pair(1))
      if cursor_y == i:
        stdscr.attron(curses.color_pair(3))
        stdscr.attron(curses.A_BOLD)
        stdscr.addstr(i, 0, display)
        if (width - l1) > 0:
          stdscr.addstr(i, l1, " " * (width - l1))
        stdscr.attroff(curses.A_BOLD)
        stdscr.attroff(curses.color_pair(3))
      else:
        stdscr.addstr(i, 0, display)
        if (width - l1) > 0:
          stdscr.addstr(i, l1, " " * (width - l1))
      stdscr.attroff(curses.color_pair(1))
    for i in range(numOfAttachments, height - 2):
      stdscr.addstr(i, 0, ' ' * width)

    # Refresh the screen
    stdscr.refresh()

    # Wait for next input
    k = stdscr.getch()

# <---

# ---> Draw message list


class State():
  '''
  Class representing the state of the program.
  TODO: implement as singleton.
  '''

  def __init__(self, **kwargs):
    # The line number of the cursor.
    self.cursor_y = kwargs.get('cursor_y', 0)
    # The scroll position.
    self.position = kwargs.get('position', 0)
    # The height of the window.
    self.height = kwargs.get('height', 0)
    # The next action to take.
    self.action = kwargs.get('action', None)
    # The selected account
    self.account = kwargs.get('account', None)
    # Any included labels.
    self.includedLabels = kwargs.get('includedLabels', ['INBOX'])
    # Any excluded labels.
    self.excludedLabels = kwargs.get('excludedLabels', ['SPAM', 'TRASH'])
    # Thread, which should be started in the background.
    self.thread = kwargs.get('thread', None)
    # Event, if not none, then we need to wait for this event.
    self.event = kwargs.get('event', None)
    # Any search terms in effect.
    self.searchTerms = kwargs.get('searchTerms', None)
    # A list of messageIds resulting from a search
    self.query = kwargs.get('query', None)
    # A list of selected messageIds.
    self.selectedMessages = kwargs.get('selectedMessages', [])
    # Visibility of labels.
    self.showLabels = kwargs.get('showLabels', False)
    # Global lock.
    self.globalLock = kwargs.get('lock', None)
    # Threading event, newMessagesArrived.
    self.newMessagesArrived = kwargs.get('newMessagesArrived', None)

  def read(self, messageId, message):
    self.action = Read(messageId, message)
    return self

  def new(self, **kwargs):
    sendTo = kwargs.get('to', None)
    subject = kwargs.get('subject', None)
    if sendTo:
      self.action = Send(type='NEW', sendTo=sendTo)
    if subject:
      try:
        self.action.addSubject(subject)
      except:
        logger.warning('Trying to add subject to non-existant send action.')

  def reply(self, messageInfo, message):
    self.action = Send(messageInfo=messageInfo, message=message, type='REPLY')
    return self

  def replyToAll(self, messageInfo, message):
    self.action = Send(messageInfo=messageInfo,
                       message=message, type='REPLY_TO_ALL')
    return self

  def forward(self, messageInfo, message, to):
    # logger.info("Trying to forward a mail to {}".format(to))
    self.action = Send(messageInfo=messageInfo,
                       message=message, type='FORWARD', sendTo=to)
    return self

  def delete(self):
    self.action = Modify(self.selectedMessages, type='DELETE')
    return self

  def markAsRead(self):
    self.action = Modify(self.selectedMessages, type='MARK_AS_READ')
    return self

  def trash(self):
    self.action = Modify(self.selectedMessages, type='TRASH')
    return self

  def viewAttachments(self):
    attachments = getAttachments(mkService(self.account),
                                 self.selectedMessages[0], self.globalLock)
    if attachments:
      self.action = ViewAttachments(attachments)
      return True
    else:
      return False

  def addQuery(self, searchTerms):
    self.position = 0
    if searchTerms:
      self.query = search(self.account, searchTerms, self.globalLock)
      self.searchTerms = searchTerms
    else:
      self.query = None
      self.searchTerms = None
    return self

  def removeQuery(self):
    self.position = 0
    self.query = None
    self.searchTerms = None
    return self

  def switchAccount(self, account):
    self.account = account
    self.position = 0
    self.removeQuery()
    return self

  def quit(self):
    self.action = Quit()
    return self

  def act(self):
    if self.action:
      # Perform action
      self.action.perform(state=self)
    else:
      logger.warning('Tried to act, but there is no action in the queue.')

# ---> mkStausBar
  def mkStatusBar(self, stdscr, height, width, numOfMessages, selectedMessage):
    '''
    Draw the status bar on the screen.

    Args:
      stdscr: The curses window object.
      height: Height of stdscr.
      width: Width of stdscr.

    Returns: None
    '''
    # user = ' \uf007 '
    # seperator = '\uf054'
    user = ' {} '.format(chr(config.user))
    seperator = chr(config.seperator)
    lUser = len(user)

    account = ' {} {} {} '.format(
        seperator,
        self.account,
        seperator)
    acLen = len(account) + lUser

    if self.searchTerms:
      st = '?[{}] {} '.format(
          self.searchTerms,
          seperator)
    else:
      st = ''
    stLen = len(st)

    mb = self.includedLabels[0]
    mbLen = len(mb)

    numOfMessagesString = ' {} [{}/{} messages]'.format(
        seperator,
        self.position + self.cursor_y + 1,
        numOfMessages)
    n = len(numOfMessagesString)

    if self.showLabels == True:
      labels = str(selectedMessage.showLabels())
    else:
      labels = ''
    labelsLen = len(labels)

    whitespace = ' {}'.format(
        ' ' * (width - mbLen - acLen - stLen - n - 1))

    stdscr.attron(curses.color_pair(5))
    stdscr.attron(curses.A_BOLD)
    stdscr.addstr(height - 2, 0, user)
    stdscr.attroff(curses.A_BOLD)

    stdscr.addstr(height - 2, lUser, account)

    stdscr.addstr(height - 2, acLen, st)

    stdscr.attron(curses.A_BOLD)
    stdscr.addstr(height - 2, acLen + stLen, mb)
    stdscr.attroff(curses.A_BOLD)

    stdscr.addstr(height - 2, acLen + stLen + mbLen, numOfMessagesString)

    stdscr.addstr(height - 2, acLen + stLen + mbLen + n, labels)

    stdscr.addstr(height - 2, acLen + stLen +
                  mbLen + n + labelsLen, whitespace)
    stdscr.attroff(curses.color_pair(5))

    if numOfMessages > 0:
      snippet = ' On {} <{}> wrote: {} {}'.format(
          selectedMessage.timeForReply(),
          selectedMessage.parseSender()[0],
          selectedMessage.snippet,
          ' ' * width
      )
      try:
        stdscr.addstr(height - 1, 0, snippet[:width - 1])
      except:
        stdscr.addstr(height - 1, 0, ' ' * (width - 1))

# <---


ACTIONS = {
    'READ_A_MAIL',
    'MARK_AS_READ',
    'REMOVE_LABELS',
    'ADD_LABELS',
    'SEND_A_MAIL',
    'FORWARD',
    'REPLY',
    'REPLY_TO_ALL',
    'DELETE',
    'DO_NOT_SEND',
}


class Action():
  '''
  Class representing actions.
  '''

  def __init__(self, action):
    self.action = action


class Read(Action):
  def __init__(self, messageId, message):
    self.messageId = messageId
    self.message = message
    Action.__init__(self, 'READ')

  def perform(self, state):
    message = self.message
    run(config.w3mArgs(), input=message, encoding='utf-8')
    # finishedUpdatingLocalDb = Event()
    state.event = Event()
    t = Thread(target=postRead,
               name="postRead",
               args=(mkService, state))
    state.thread = t
    # state.markAsRead().act()
    return state


class Modify(Action):

  def __init__(self, selectedMessages, type):
    self.selectedMessages = selectedMessages
    self.type = type
    Action.__init__(self, 'MODIFY')

  def perform(self, state):

    ms = self.selectedMessages
    lock = state.globalLock

    if self.type == 'DELETE':
      logger.info('Modify local label DB: + "TRASH", - "UNREAD".')
      data = {'action': 'ADD_LABELS',
              'labels': [(m.messageId, ['TRASH']) for m in ms]}
      sendToServer(data,lock)
      data = {'action': 'REMOVE_LABELS',
              'labels': [(m.messageId, ['UNREAD']) for m in ms]}
      sendToServer(data,lock)

    elif self.type == 'MARK_AS_READ':
      # Remove 'UNREAD' labels.
      logger.info('Modify local label DB: - "UNREAD".')
      data = {'action': 'REMOVE_LABELS',
              'labels': [(m.messageId, ['UNREAD']) for m in ms]}
      sendToServer(data,lock)

    elif self.type == 'TRASH':
      # Remove 'UNREAD' labels.
      logger.info('Modify local label DB: + "TRASH".')
      data = {'action': 'ADD_LABELS',
              'labels': [(m.messageId, ['TRASH']) for m in ms]}
      sendToServer(data,lock)

    t = Thread(target=postDelete, name='DELETE',
               args=(state, mkService, [m.messageId for m in ms]))
    t.start()


class Send(Action):
  def __init__(self, **kwargs):
    self.messageInfo = kwargs.get('messageInfo', None)
    self.message = kwargs.get('message', None)
    self.sendTo = kwargs.get('sendTo', None)
    self.type = kwargs.get('type', None)
    Action.__init__(self, 'SEND')

  def addSubject(self, subject):
    self.subject = subject

  def confirm(self, account):
    curses.wrapper(lambda x: drawConfirmationScreen(x, account, self))

  def perform(self, state):
    if self.type == 'NEW':
      draftId = str(uuid4())
      messageInfo = None
      input = None

    elif self.type in ['REPLY', 'FORWARD', 'REPLY_TO_ALL']:
      messageInfo = self.messageInfo
      message, draftId = self.message, messageInfo.messageId

      self.subject = mkSubject(messageInfo, self.type)

      if self.type in ['REPLY', 'REPLY_TO_ALL']:
        self.sendTo = mkTo(state.account, messageInfo, self.type)
      # elif self.type in ['FORWARD']:
        # logger.info('self.sendTo = {}'.format(self.sendTo))

      # First format the message.
      formatedMessage = run(config.w3mArgs(), input=message,
                            encoding='utf-8', stdout=PIPE)

      # Rewrite mkReplyInfo/mkForwardInfo...
      input = addInfo(messageInfo, formatedMessage, self.type)

    # Check that tmp file doesn't exist and remove if it does.
    if os.path.exists(os.path.join(config.tmpDir, draftId)):
      os.remove(os.path.join(config.tmpDir, draftId))

    # Open the formated message in vim.
    run(config.vimArgs(draftId), input=input, encoding='utf-8')

    # Check that tmp file exists ready to be sent.
    if os.path.exists(os.path.join(config.tmpDir, draftId)):
      type = self.type if self.type in [
          'NEW', 'FORWARD'] else 'REPLY'

      # Run confirmation loop.
      self.confirm(state.account)

      # Make the thread.
      if self.send == 'SEND':
        finishedUpdatingLocalDb = Event()
        t = Thread(target=postSend, name="postSend",
                   args=(mkService, finishedUpdatingLocalDb,
                         state.account, draftId),
                   kwargs={'type': type,
                           'attachments': self.attachments,
                           'to': self.sendTo,
                           'cc': self.cc,
                           'bcc': self.bcc,
                           'subject': self.subject,
                           'globalLock': state.globalLock,
                           'messageInfo': messageInfo})
        state.thread = t
        state.event = finishedUpdatingLocalDb
      elif self.send == 'DO_NOT_SEND':
        state.thread = None
    else:
      state.thread = None
    return state


class ViewAttachments(Action):
  def __init__(self, attachments):
    self.attachments = attachments
    Action.__init__(self, 'VIEW_ATTACHMENTS')

  def perform(self, state):
    # do something here and return something depending on if succesful or not
    curses.wrapper(lambda x: drawAttachments(
        x, state.account, state, self.attachments))


class Quit(Action):
  def __init__(self):
    Action.__init__(self, 'QUIT')

  def perform(self):
    pass


def drawMessages(stdscr, state, accountSwitcher, eventQue):
  '''
  Loop to draw main screen of the program. A list of messages, which
  depends on the programs state.

  Args:
    stdscr: The curses window object.
    state: The state of the program.
  '''
  # Start colors in curses
  curses.curs_set(0)
  initColors()

  # Check if there are threads that need starting and
  # wait for them to complete certain tasks if necessary.
  if state.thread:
    state.thread.start()
    logger.info('Starting thread: {}.'.format(state.thread.name))
    state.thread = None
    if state.event:
      state.event.wait()

  # Clear and refresh the screen for a blank canvas
  stdscr.clear()
  stdscr.refresh()
  height, width = stdscr.getmaxyx()
  state.height = height

  # Update the list of messages.
  messages = getMessages(state, returnCount=False)
  # if state.thread:
  #   if state.thread.name == "postRead":
  #     logger.info('Getting new message list.')
  numOfMessages = getMessages(state, returnCount=True)
  selectedMessage = messages[state.cursor_y]
  # logger.info('Labels: {}'.format([l.label for l in messages[0].labels]))

  # Loop where k is the last character pressed
  k = 0
  while 1:

    height, width = stdscr.getmaxyx()
    state.height = height
    # logger.info('k is {}'.format(k))

    if k == curses.KEY_RESIZE:
      height, width = stdscr.getmaxyx()
      state.height = height
      stdscr.addstr(height - 1, 0, ' ' * (width - 1))

    elif k == ord(' '):
      # Select mail
      if selectedMessage not in state.selectedMessages:
        state.selectedMessages.append(messages[state.cursor_y])
      else:
        state.selectedMessages.remove(messages[state.cursor_y])
      k = curses.KEY_DOWN
      continue

    elif k in [curses.KEY_DOWN, ord('j')]:
      state.cursor_y = state.cursor_y + 1
      if state.cursor_y == height - 2:
        state.position = min(state.position + 1, numOfMessages - height + 2)
        messages = getMessages(state, returnCount=False)

    elif k in [curses.KEY_UP, ord('k')]:
      state.cursor_y = state.cursor_y - 1
      if state.cursor_y == -1:
        state.position = max(state.position - 1, 0)
        messages = getMessages(state, returnCount=False)

    elif k in [curses.KEY_NPAGE, 4]:
      state.position = min(
          state.position + height - 7, max(numOfMessages - height + 2, 0), numOfMessages - 1)
      messages = getMessages(state, returnCount=False)

    elif k in [curses.KEY_PPAGE, 21]:
      state.position = max(state.position - height + 7, 0)
      messages = getMessages(state, returnCount=False)

    elif k == ord('M'):
      state.cursor_y = (height - 2)//2

    elif k == ord('H'):
      state.cursor_y = 0

    elif k == ord('L'):
      state.cursor_y = height - 2

    elif k == ord('\n'):
      # Read a message.
      message = readMessage(mkService(state.account),
                            selectedMessage.messageId)
      state.selectedMessages = [selectedMessage]
      state.read(selectedMessage.messageId, message)
      return state

    elif k == ord('r'):
      # Reply to message.
      message = readMessage(mkService(state.account),
                            selectedMessage.messageId)
      state.reply(selectedMessage, message)
      stdscr.addstr(height - 1, 0, ' ' * (width - 1))
      curses.curs_set(0)
      return state

    elif k == ord('f'):
      # Forward message.
      to = chooseAddress(state.account)
      if to:
        message = readMessage(mkService(state.account),
                              selectedMessage.messageId)
        state.forward(selectedMessage, message, to)
        stdscr.addstr(height - 1, 0, ' ' * (width - 1))
        curses.curs_set(0)
        return state

    elif k == ord('a'):
      # Reply to group/all.
      message = readMessage(mkService(state.account),
                            selectedMessage.messageId)
      state.replyToAll(selectedMessage, message)
      stdscr.addstr(height - 1, 0, ' ' * (width - 1))
      curses.curs_set(0)
      return state

    elif k == ord('m'):
      # Make a new message.
      to = chooseAddress(state.account)
      stdscr.clear()
      stdscr.refresh()
      if to:
        # This is a hack to force screen to be redrawn before prompting
        # for a subject.
        state.new(to=to)
        # state = {'to': to}
        k = 'm0'
        continue

    elif k == ord('g'):
      k = stdscr.getch()
      if k == ord('g'):
        state.position = 0
        state.cursor_y = 0
        messages = getMessages(state, returnCount=False)

    elif k == ord('G'):
      state.cursor_y = height - 2
      state.position = numOfMessages - height + 2
      messages = getMessages(state, returnCount=False)

    elif k == ord('d'):
      # Delete message in varous ways.
      k = stdscr.getch()
      if len(state.selectedMessages) == 0:
        state.selectedMessages.append(selectedMessage)
      if k == ord('d'):
        # Delete message (add 'TRASH' label and
        # remove 'UNREAD' label if present).
        state.delete().act()

      elif k == ord('r'):
        # Mark as read. Remove unread label but leave in INBOX.
        state.markAsRead().act()

      elif k == ord('t'):
        # Add to TRASH but don't read.
        state.trash().act()

      messages = getMessages(state, returnCount=False, afterAction={
          'action': state.action.type,
          'messageIds': [m.messageId for m in state.selectedMessages]
      })
      numOfMessages = getMessages(state, returnCount=True)
      state.selectedMessages = []

    elif k == ord('l'):
      # Various label related features.
      k = stdscr.getch()
      if k == ord('l'):
        # Show the labels on the highlighted message.
        showLabels = not showLabels
      else:
        if k == ord('u'):
          # Show unread messages.
          state.excludedLabels = ['SPAM', 'TRASH']
          state.includedLabels = ['UNREAD']
        elif k == ord('i'):
          # Show messages in the inbox (default)
          state.excludedLabels = ['SPAM', 'TRASH']
          state.includedLabels = ['INBOX']
        elif k == ord('s'):
          # Show sent messages.
          state.excludedLabels = ['TRASH']
          state.includedLabels = ['SENT']
        elif k == ord('t'):
          # Show the trash.
          state.excludedLabels = ['SPAM']
          state.includedLabels = ['TRASH']
        state.position = 0
        messages = getMessages(state, returnCount=False)
        numOfMessages = getMessages(state, returnCount=True)

    elif k == ord('/'):
      # Do a search.
      searchTerms = getInput(stdscr, "Enter search terms: ", height, width)
      if searchTerms:
        state.addQuery(searchTerms)
        messages = getMessages(state, returnCount=False)
        numOfMessages = getMessages(state, returnCount=True)
      curses.curs_set(0)

    elif k == ord('c'):
      # Clear search terms.
      if state.searchTerms:
        state.addQuery(None)
        messages = getMessages(state, returnCount=False)
        numOfMessages = getMessages(state, returnCount=True)
      curses.curs_set(0)

    elif k == ord('v'):
      # View attachments.
      # attachments = getAttachments(mkService(state.account), selectedHeader)
      state.selectedMessages = [selectedMessage]
      attachments = state.viewAttachments()
      messages = getMessages(state, returnCount=False)
      if attachments:
        return state

    elif k == ord('\t'):
      # Toggle between accounts.
      state.switchAccount(next(accountSwitcher))
      messages = getMessages(state, returnCount=False)
      numOfMessages = getMessages(state, returnCount=True)

    elif k == ord('q'):
      # Quit.
      state.quit()
      return state

    if numOfMessages == 0:
      for i in range(height - 2):
        stdscr.addstr(i, 0, ' ' * width)
      noMessages = "No messages found! Press 'c' to go back."
      stdscr.addstr(height - 1, 0, noMessages)
      stdscr.addstr(height - 1, len(noMessages), " " *
                    (width - 1 - len(noMessages)))
    else:
      # Update line number.
      state.cursor_y = max(0, state.cursor_y)
      state.cursor_y = min(height-3, state.cursor_y, numOfMessages - 1)
      # Update selected message.
      selectedMessage = messages[state.cursor_y]
      # state.selectedMessage = selectedMessage

      for i, h in enumerate(messages[:height - 2]):
        display = h.display(15, width)
        l1 = len(display)
        stdscr.attron(curses.color_pair(1))
        if state.cursor_y == i and h in state.selectedMessages:
          stdscr.attron(curses.color_pair(4))
          stdscr.attron(curses.A_BOLD)
          stdscr.addstr(i, 0, display)
          if (width - l1) > 0:
            stdscr.addstr(i, l1, " " * (width - l1))
          stdscr.attroff(curses.A_BOLD)
          stdscr.attroff(curses.color_pair(4))
        elif state.cursor_y == i:
          stdscr.attron(curses.color_pair(3))
          stdscr.attron(curses.A_BOLD)
          stdscr.addstr(i, 0, display)
          if (width - l1) > 0:
            stdscr.addstr(i, l1, " " * (width - l1))
          stdscr.attroff(curses.A_BOLD)
          stdscr.attroff(curses.color_pair(3))
        elif h in state.selectedMessages:
          stdscr.attron(curses.color_pair(2))
          stdscr.attron(curses.A_BOLD)
          stdscr.addstr(i, 0, display)
          if (width - l1) > 0:
            stdscr.addstr(i, l1, " " * (width - l1))
          stdscr.attroff(curses.A_BOLD)
          stdscr.attroff(curses.color_pair(2))
        else:
          stdscr.addstr(i, 0, display)
          if (width - l1) > 0:
            stdscr.addstr(i, l1, " " * (width - l1))
        stdscr.attroff(curses.color_pair(1))
      if numOfMessages < height - 2:
        for i in range(numOfMessages, height - 1):
          stdscr.addstr(i, 0, " " * width)

    state.mkStatusBar(stdscr, height, width, numOfMessages, selectedMessage)

    if state.action and state.action.action == 'VIEW_ATTACHMENTS' and state.action.saved == 'SAVED_ATTACHMENT':
      # if state['action'] == 'SAVED_ATTACHMENT':
      putMessage(stdscr, height, width, "Attachment saved successfully.")
      state.action = None
      state.selectedMessages = []

    if k == 'm0':
      # Follow up on composing a new message. Ask for the subject.
      subject = getInput(stdscr, "Subject: ", height, width)
      if subject:
        state.new(subject=subject)
        return state
      else:
        state.action = None
        # state.pop('action', None)
      stdscr.addstr(height - 1, 0, ' ' * (width - 1))
      curses.curs_set(0)
    # Refresh the screen
    stdscr.refresh()

    # stdscr.nodelay(True)
    t = Thread(target=waitForKey,args=(stdscr,eventQue),
               daemon=True)
    t.start()

    e = eventQue.get()

    if e['event'] == 'KeyPress':
      k = e['value']
    elif e['event'] == 'NewMsg':
      logger.info('Redrawing message list')
      state.cursor_y = 0
      state.position = 0
      messages = getMessages(state,returnCount=False)
      k = None

# <---

def waitForKey(stdscr,eventQue):
  # logger.info('waiting for keypress')
  k = stdscr.getch()
  # logger.info('keypress not registering')
  eventQue.put({'event': 'KeyPress','value':k})

# <---

# ---> Post and Preprocessing functions

# ---> Postprocessing


def postDelete(state, service, messageIds):
  '''
  Modify labels on remote db after deleting.

  Args:
    state: The state of the program.
    service: The google API sevice object.
    messagedIds: List of message ids which were
    deleted.

  Returns:
    None
  '''
  # Remove unread label from Google servers.
  logger.info('Modify labels in remote DB: + "TRASH", - "UNREAD".')
  if state.action.type == 'DELETE':
    body = {'ids': messageIds,
            'removeLabelIds': ['UNREAD'],
            'addLabelIds': ['TRASH']}
  elif state.action.type == 'MARK_AS_READ':
    body = {'ids': messageIds,
            'removeLabelIds': ['UNREAD'],
            'addLabelIds': []}
  elif state.action.type == 'TRASH':
    body = {'ids': messageIds,
            'removeLabelIds': [],
            'addLabelIds': ['TRASH']}
  account = state.account
  try:
    service(account).users().messages().batchModify(
        userId='me', body=body).execute()
  except errors.Error as e:
    logger.debug(e)


def postSend(service, event, account, draftId, **kwargs):
  '''
  Actually send the message.
  Gets run in a thread while the mainloop restarts.

  Args:
    service: API service object.
    event: threading.Event object - set it to true once local db has updated.
    sender: Whoever is sending email.
    draftId: The id corresponding to the tmp file of the message text.

  Keyword args:
    type: One of NEW, REPLY, FORWARD.
    attachments: If present a list of files to attach.
    to: Who to send the mail to.
    cc: Carbon copies sent here.
    bcc: Blind carbon copies.
    subject: Subject of the mail.
    messageInfo: If type is reply or foreward, this must be present.

  Returns: None
  '''

  try:
    with open(os.path.join(config.tmpDir, draftId), 'r') as f:
      message = createMessage(account, f.read(), **kwargs)
    # Send the message.
    message = sendMessage(service(account), account, message)
    logger.info('Sent a message.')
    # Mark it as read if it wasn't already.
    # if kwargs['type'] in ['REPLY', 'FORWARD']:
    # messageInfo = kwargs['messageInfo']
    # postRead(account, event, service, messageInfo.messageId)
    # Clean up.
    os.remove(os.path.join(config.tmpDir, draftId))
    # Add message to local db.
    data = {'action': 'ADD_MESSAGES',
            'account': account,
            'messageIds': [message['id']]}
    lock = kwargs.get('globalLock',None)
    sendToServer(data,lock)
    # logger.info('data sent to server')
    event.set()
    # make this message more useful!
    # logger.info('Email sent successfully.')
  except Exception as e:
    # Something went wrong.
    logger.warning(e)


# def postRead(account, event, service, messageId, lock):
def postRead(service, state):
  '''
  Mark message as read after reading.

  Args:
    account: The currently selected account.
    event: threading.Event object - set it to true once local db has updated.
    service: The google API service object.
    messageId: The id of the highlighted message.

  Returns:
    None
  '''
  # Mark message as read.
  # Remove 'UNREAD' label from local storage.
  # logger.info('Removing label from local storage.')
  account = state.account
  messageId = state.selectedMessages[0].messageId
  lock = state.globalLock
  data = {'action': 'REMOVE_LABELS',
          'labels': [(messageId, ['UNREAD'])],
          'account': account,
          'query': state.query,
          'position': state.position,
          'height': state.height,
          'excludedLabels': state.excludedLabels,
          'includedLabels': state.includedLabels,
          'afterAction': 'MARK_AS_READ'}
  sendToServer(data,lock)
  # Set event to true so that main thread can continue.
  state.event.set()
  state.selectedMessages = []
  # Remove unread label from Google servers.
  # logger.info('Removing label from Google servers.')
  body = {'removeLabelIds': ['UNREAD'], 'addLabelIds': []}
  try:
    service(account).users().messages().modify(userId='me', id=messageId,
                                               body=body).execute()
  except errors.Error as e:
    logger.debug(e)

# <---

# ---> Preprocessing


def readMessage(service, messageId):
  '''
  Retrive a message from the server ready for reading.

  Args:
    service: The google API service object.
    messageId: The id of the message to retrive.

  Reurns:
    The message as a string ready for w3m.
  '''
  msg = service.users().messages().get(
      userId='me', id=messageId, format='raw').execute()
  msg = msg['raw']
  msg = base64.urlsafe_b64decode(msg).decode('utf-8')
  msg = email.message_from_string(msg, policy=email.policy.default)
  return msg.get_body(('html', 'plain',)).get_content()


'''
def preSend(sender, state):
  Collect and organise various information
  prior to actually sending the mail.

  Args:
    sender: Who will send the email.
    state: A dictonary with various information
      action: One of NEW, REPLY, FORWARD, REPLYTOALL.
      to: Who to send the mail to. Only needed if action is NEW or FORWARD.
      subject: Only if action is NEW.
      message: Present unless action is NEW. It is the message being replied
        to/forwarded.
      header: If action is not NEW, this must be present.

  Returns:
    state: Same as input, and unless the user had a change of heart
      a thread key.
      thread: Thread to be run later.
  if state['action'] == 'NEW':
    draftId = str(uuid4())
    header = None
    input = None

  elif state['action'] in ['REPLY', 'FORWARD', 'REPLYTOALL']:
    header = state['header']
    message, draftId = state['message'], header.messageId

    state['subject'] = mkSubject(header, state['action'])

    if state['action'] in ['REPLY', 'REPLYTOALL']:
      state['to'] = mkTo(sender, header, state['action'])

    # First format the message.
    formatedMessage = run(config.w3mArgs(), input=message,
                          encoding='utf-8', stdout=PIPE)

    # Rewrite mkReplyInfo/mkForwardInfo...
    input = addInfo(header, formatedMessage, state['action'])

  # Check that tmp file doesn't exist and remove if it does.
  if os.path.exists(os.path.join(config.tmpDir, draftId)):
    os.remove(os.path.join(config.tmpDir, draftId))

  # Open the formated message in vim.
  run(config.vimArgs(draftId), input=input, encoding='utf-8')

  # Check that tmp file exists ready to be sent.
  if os.path.exists(os.path.join(config.tmpDir, draftId)):
    type = state['action'] if state['action'] in [
        'NEW', 'FORWARD'] else 'REPLY'

    # Run confirmation loop.
    state = curses.wrapper(lambda x: drawConfirmationScreen(x, sender, state))

    # Make the thread.
    if state['action'] == 'SEND':
      finishedUpdatingLocalDb = Event()
      t = Thread(target=postSend,
                 args=(mkService, finishedUpdatingLocalDb, sender, draftId),
                 kwargs={'type': type,
                         'attachments': state['attachments'],
                         'to': state['to'],
                         'cc': state['cc'],
                         'bcc': state['bcc'],
                         'subject': state['subject'],
                         'header': header})
      state['thread'] = t
      state['event'] = finishedUpdatingLocalDb
    elif state['action'] == 'DO_NOT_SEND':
      state.pop('thread', None)
  else:
    state.pop('thread', None)
  return state

'''


def addInfo(header, formatedMessage, type):
  '''
  Add a line containing infomation, and indent
  replies and forwarded messages.

  Args:
    header: the MessageInfo object of the mail being
    replied to or forwarded.
    formatedMessage: The formatedMessage as output by w3m
    as stdout.
    type: One of 'REPLY', 'REPLY_TO_ALL', 'FORWARD'.
  Returns:
    A String consisting of the new message.
  '''
  lines = []
  for line in formatedMessage.stdout:
    if line == '\n':
      lines.append(line + '> ')
    else:
      lines.append(line)

  if type in ['REPLY', 'REPLY_TO_ALL']:
    info = 'On {} <{}> wrote:'.format(
        header.timeForReply(),
        header.parseSender()[0])
  elif type == 'FORWARD':
    sender = header.parseSender()
    info = ' ----- Forwarded message from {} <{}> -----'.format(
        sender[1],
        sender[0])

  lines = ['\n\n', info, '\n\n>'] + lines
  return ''.join(lines)

# <---

# ---> Processing Attachments


def saveAttachment(service, attachment):
  '''
  Save an attachement.

  Args:
    service: The google API service object.
    attachment: The attachment object to be saved.

  Returns:
    None
  '''
  messageId = attachment.messageId
  try:
    message = service.users().messages().get(userId='me',
                                             id=messageId).execute()

    for part in message['payload']['parts']:
      if part['filename']:
        if part['partId'] == attachment.partId:

          if 'data' in part['body']:
            data = base64.urlsafe_b64decode(part['body']['data']
                                            .encode('utf-8'))
          else:
            attachmentId = part['body']['attachmentId']
            a = service.users().messages().attachments()\
                .get(userId='me', id=attachmentId, messageId=messageId)\
                .execute()
            data = base64.urlsafe_b64decode(a['data'].
                                            encode('utf-8'))

          path = os.path.join(config.dlDir, part['filename'])

          with open(path, 'wb') as f:
            f.write(data)

  except errors.Error as e:
    logger.warning('An error occured while tring to save an attachment.')
    logger.warning('An error occurred: {}'.format(e))


def getAttachments(service, header, lock):
  '''
  Get a list of attachements, if any, from the highlighted message.

  Args:
    service: The google API service.
    header: The MessageInfo object corresponding to the highlighted
    message.

  Returns:
    List of attachment objects or None if there are no attachments.
  '''
  messageId = header.messageId
  attachments = []
  data = {'action': 'GET_QUERY',
          'class': Attachments,
          'messageIds': [messageId]}
  q = sendToServer(data,lock)
  # q = s.query(Attachments).filter(Attachments.messageId == messageId)
  if len(q) > 0:
    # Attachment info exists in db - use this.
    attachments = q
  else:
    # Fetch attachment info.
    try:
      message = service.users().messages().get(userId='me',
                                               id=messageId).execute()

      for part in message['payload']['parts']:
        if part['filename']:
          a = Attachments(
              messageId,
              part['partId'],
              part['filename'],
              part['mimeType'],
              part['body']['size'])
          # s.add(a)
          data = {'action': 'ADD_ATTACHMENT',
                  'attachment': a}
          sendToServer(data,lock)
          attachments.append(a)

    except errors.HttpError as e:
      logger.debug('An error occurred: {}'.format(e))
  if len(attachments) == 0:
    # In the future this can probably happen automagically.
    logger.info('Attempting to remove false positive attachment signal')
    data = {'action': 'REMOVE_FALSE_ATTACMENTS',
            'messageId': messageId}
    sendToServer(data,lock)
    # q = s.query(MessageInfo).filter(MessageInfo.messageId == messageId)\
    #     .update({MessageInfo.hasAttachments: False},
    #             synchronize_session='evaluate')
    # s.commit()
    logger.info('Signal removed.')
  else:
    return attachments
# <---

# <---

# ---> Searching and filtering.


def search(account, searchTerms, lock):
  '''
  Do a search.

  Args:
    account: Currently selected account.
    searchTerms: What to search for.

  Returns:
    query consiting of MessageInfo objects.
  '''
  service = mkService(account)
  messageIds = [m['id'] for m in
                listMessagesMatchingQuery(service, 'me', query=searchTerms)]
  data = {'action': 'ADD_MESSAGES',
          'account': account,
          'messageIds': messageIds}
  sendToServer(data,lock)
  data = {'action': 'GET_QUERY',
          'class': MessageInfo,
          'messageIds': messageIds}
  return [q.messageId for q in sendToServer(data,lock)]


def listFiles(directory):
  '''
  List the files in directory as a
  generator.
  '''
  for root, _, files in os.walk(directory):
    for f in files:
      yield os.path.join(root, f)


def chooseAttachment(x): return fzf(
    'Choose an attachment: ', listFiles(x))


def chooseAddress(x): return fzf(
    'To: ', AddressBook.addressList(x))


'''
# Wrapper for fzf to dump and restore window state before and after.
def fzfWrapper (stdscr, prompt, iterable):
  stdscr.refresh()
  with open(TMPDIR_WINDOW + 'stdscr.window','wb') as f:
    stdscr.putwin(f)
  out = fzf(prompt,iterable)
  stdscr.clear()
  with open(TMPDIR_WINDOW + 'stdscr.window','rb') as f:
    stdscr = curses.getwin(f)
  stdscr.refresh()
  return out
'''


def fzf(prompt, iterable):
  '''
  Use fzf to fuzzy search an iterable.
  Stole this from http:github.com/dahlia/iterfzf.

  Args:
    prompt: Prompt to show the user.
    iterable: iterable to search.
  '''
  proc = None
  for line in iterable:
    line = re.sub('\n', '', line)
    line = re.sub('\r', '', line)
    if proc is None:
      proc = Popen(
          config.fzfArgs(prompt),
          stdin=PIPE,
          stdout=PIPE,
          stderr=None
      )
      stdin = proc.stdin
    line = line.encode('utf-8')
    try:
      stdin.write(line + b'\n')
      stdin.flush()
    except IOError as e:
      if e.errno != errno.EPIPE and errno.EPIPE != 32:
        raise
      break
  if proc is None or proc.wait() not in [0, 1]:
    return None
  try:
    stdin.close()
  except IOError as e:
    if e.errno != errno.EPIPE and errno.EPIPE != 32:
      raise
  stdout = proc.stdout
  def decode(t): return t.decode('utf-8')
  output = [decode(ln.strip(b'\r\n\0')) for ln in iter(stdout.readline, b'')]
  try:
    return output[-1]
  except IndexError:
    return None

# <---

# ---> Main


def sendToServer(data, lock):
  with lock:
    host = socket.gethostname()
    port = config.port
    sock = socket.socket()
    bufferSize = 1024

    try:
      sock.connect((host, port))
      pickledData = pickle.dumps(data)
      sizeOfPickle = len(pickledData)
      numOfChunks = sizeOfPickle//bufferSize + 1
      sock.send(sizeOfPickle.to_bytes(4, 'big'))

      for i in range(numOfChunks):
        sock.send(pickledData[i*bufferSize: (i+1)*bufferSize])

      sizeOfResponse = int.from_bytes(sock.recv(4), 'big')
      response = b''

      while len(response) < sizeOfResponse:
        response += sock.recv(bufferSize)

      unpickledResponse = pickle.loads(response)
      sock.close()
      return unpickledResponse

    except Exception as e:
      logger.warning('Could not connect to pmailServer')
      print('Could not connect on port!')
      logger.debug(e)
      sys.exit()


def setEscDelay():
  '''
  This is needed so that when pressing ESCAPE to exit from the getInput
  function there is no noticable delay.
  '''
  os.environ.setdefault('ESCDELAY', '25')


def getMessages(state, **kwargs):
  returnCount = kwargs.get('returnCount', False)
  afterAction = kwargs.get('afterAction', None)
  data = {'action': 'GET_MESSAGES',
          'account': state.account,
          'query': state.query,
          'position': state.position,
          'height': state.height,
          'excludedLabels': state.excludedLabels,
          'includedLabels': state.includedLabels,
          'count': returnCount,
          'afterAction': afterAction}
  return sendToServer(data, state.globalLock)


def mainLoop(lock,accountSwitcher, eventQue):
  '''
  Main loop of the program. This loop processes the state
  based on its 'action' and launches various subloops.
  '''
  logger.info('Starting main loop...')
  account = next(accountSwitcher)
  # Initialise the state.
  state = State(account=account,
                lock=lock)
  while True:

    # Run inner loop and update the state when it exits.
    state = curses.wrapper(
      lambda x: drawMessages(x, state, accountSwitcher, eventQue))
    if type(state.action).__name__ == 'Quit':
      break
    else:
      state.act()

def checkForNewMessages(lock, eventQue):
  data = {'action':'CHECK_FOR_NEW_MESSAGES',
          }
  logger.info('Asking server if anything new showed up.')
  while 1:
    response = sendToServer(data, lock)
    if response == 'newMessagesArrived':
      logger.info('New message detected.')
      # newMessagesArrived.clear()
      eventQue.put({'event': 'NewMsg'})
    sleep(10)

# if __name__ == '__main__':
def start():
  setEscDelay()
  accountSwitcher = cycle(config.listAccounts())
  logger.setLevel(logging.DEBUG)
  # TODO: Not using newMessagesArrived event any more!
  # newMessagesArrived = Event()
  lock = Lock()
  eventQue = Queue()
  t1 = Thread(target=mainLoop,args=(lock,accountSwitcher,eventQue,))
  t2 = Thread(target=checkForNewMessages,args=(lock,eventQue,),
              daemon=True)
  # msgListener = Thread(target=waitForNewMessages,
  #        args=(eventQue,newMessagesArrived))
  t1.start()
  t2.start()
  # msgListener.start()
  # mainLoop()

  # Clean up tmp files.
  for f in os.listdir(config.tmpDir):
    os.remove(os.path.join(config.tmpDir, f))
# <---

"""
vim:foldmethod=marker foldmarker=--->,<---
"""



#!/usr/bin/python

# ---> Imports
import re
import base64
import mimetypes
import os

from apiclient import errors
from pmail.common import logger, config
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
# <---

# ---> Send a message


def sendMessage(service, userId, message):
  """Send an email message.

  Args:
    service: Authorized Gmail API service instance.
    user_id: User's email address. The special value "me"
    can be used to indicate the authenticated user.
    message: Message to be sent.

  Returns:
    Sent Message.
  """
  try:
    message = (service.users().messages().
               send(userId=userId, body=message).execute())
    # print('Message Id: %s' % message['id'])
    return message
  except errors.Error as e:
    logger.warning(e)

# <---

# ---> Create a message


def createMessage(sender, messageText, **kwargs):
  '''
  Create a message.

  Args:
    sender: Email address of the sender.
    messageText: The text to appear in the body of the mail.

  Keyword args:
    type: One of NEW, REPLY, FORWARD.
    attachments: If present a list of files to attach.
    to: Who to send the mail to. 
    cc: Carbon copies sent here.
    bcc: Blind carbon copies.
    subject: Subject line of the mail. 
    header: If type is reply or foreward, this must be present.

  Returns: Base64-encoded email message.
  '''
  type = kwargs.get('type', 'NEW')
  attachments = kwargs.get('attachments', None)
  to = kwargs.get('to', None)
  cc = kwargs.get('cc', None)
  bcc = kwargs.get('bcc', None)
  subject = kwargs.get('subject', None)
  header = kwargs.get('messageInfo', None)

  if attachments:
    message = MIMEMultipart()
  else:
    message = MIMEText(messageText, _charset='utf-8')

  message['From'] = '{} <{}>'.format(config.getName(sender), sender)
  message['To'] = to
  if cc:
    message['Cc'] = cc
  if bcc:
    message['Bcc'] = bcc
  message['Subject'] = subject

  if type == 'REPLY':
    if header.references:
      references = header.externalId + ' ' + header.references
    elif header.inReplyTo:
      references = header.externalId + ' ' + header.inReplyTo
    else:
      references = header.externalId
    message['In-Reply-To'] = header.externalId
    message['References'] = references
  if attachments:
    msg = MIMEText(messageText, _charset='utf-8')
    message.attach(msg)

    for file in attachments:
      content_type, encoding = mimetypes.guess_type(file)

      if content_type is None or encoding is not None:
        content_type = 'application/octet-stream'
      main_type, sub_type = content_type.split('/', 1)
      if main_type == 'text':
        with open(file, 'rb') as f:
          msg = MIMEText(f.read(), _subtype=sub_type)
      elif main_type == 'image':
        with open(file, 'rb') as f:
          msg = MIMEImage(f.read(), _subtype=sub_type)
      elif main_type == 'audio':
        with open(file, 'rb') as f:
          msg = MIMEAudio(f.read(), _subtype=sub_type)
      else:
        with open(file, 'rb') as f:
          msg = MIMEBase(main_type, sub_type)
          msg.set_payload(f.read())

      msg.add_header('Content-Disposition', 'attachment',
                     filename=os.path.basename(file))
      message.attach(msg)

  return {'raw': base64.urlsafe_b64encode(
      message.as_string().encode()).decode()}


def mkSubject(header, type):
  '''
  If sending a reply or forward then make the subject.

  Args:
    header: The MessageInfo object of the message being replied 
    or forewarded.
    type: One of REPLY, REPLY_TO_ALL, FORWARD.

  Returns:
    A string with the subject.
  '''
  if type in ['REPLY', 'REPLY_TO_ALL']:
    if (header.subject)[:3] not in ['re:', 'Re:']:
      subject = 'Re: ' + header.subject
    else:
      subject = header.subject
  elif type == 'FORWARD':
    subject = 'Fwd: ' + header.subject
  return subject


def mkTo(sender, header, type):
  '''
  If sending a reply figure out who to reply to.
  
  Args:
    sender: Whoever is sending mail.
    header: The MessageInfo object of the mail being replied to.
    type: One of REPLY or REPLY_TO_ALL.
  '''
  if type == 'REPLY':
    to = header.sender
  elif type == 'REPLY_TO_ALL':
    all = filter(lambda x: not re.search(sender, x),
                 (header.recipients).split(','))
    to = header.sender
    for a in all:
      to += (', ' + a)
  return to

# <---

"""
vim:foldmethod=marker foldmarker=--->,<---
"""



#!/usr/bin/python

# ---> Imports
from __future__ import print_function
import sys
from datetime import datetime, timezone
import os
import os.path
import re
import pickle
import logging
import yaml
from yaml import Loader
from sqlalchemy import create_engine, desc, UniqueConstraint
from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import sessionmaker, relationship, scoped_session
from sqlalchemy.ext.declarative import declarative_base
from googleapiclient.discovery import build
from googleapiclient.discovery_cache.base import Cache
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
# <---

# ---> Initial definitions

WORKING_DIR = os.path.dirname(os.path.abspath(__file__))

class Config():
  # def __init__(self,config_path=os.path.join(WORKING_DIR,'config.yaml')):
  def __init__(self,**kwargs):
    try:
      home = os.environ['HOME']
    except KeyError as e:
      logger.debug(e)
      logger.warning('HOME not set.')
      sys.exit()
    pmailDir = '.local/share/pmail'
    configDir = '.config/pmail'
    if 'config_path' not in kwargs:
      config_path_0 = os.path.join(home,configDir,'config.yaml')
      config_path_1 = os.path.join(WORKING_DIR,'config.yaml')
      if os.path.exists(config_path_0):
        config_path = config_path_0
      elif os.path.exists(config_path_1):
        config_path = config_path_1
      else:
        msg = "Could not find config file 'config.yaml'." +\
              "Pmail looked in {} and {}."\
              .format(config_path_0,config_path_1)
        logger.warning(msg)
        print(msg)
        sys.exit()
    else:
      config_path = kwargs.get('config_path')
    
    with open(config_path, 'r', encoding='utf-8') as f:
      y = yaml.load(f, Loader=Loader)
    b = y['behaviour']
    self.syncFrom = b['sync_from']
    self.editor = b['editor']
    self.pager = b['pager']
    self.picker = b['picker']
    # self.tmpDir = b['tmp_directory']
    tmpPath = os.path.join(home,pmailDir,'tmp')
    if not os.path.exists(tmpPath):
      os.makedirs(tmpPath)
    self.tmpDir = b.get('tmp_directory',tmpPath)
    # self.pickleDir = b['pickle_directory']
    picklePath = os.path.join(home,pmailDir,'pickles')
    if not os.path.exists(picklePath):
      os.makedirs(picklePath)
    self.pickleDir = b.get('pickle_directory', picklePath) 
    # self.dlDir = b['download_directory']
    dlPath = os.path.join(home,'Downloads')
    if not os.path.exists(dlPath):
      os.makedirs(dlPath)
    self.dlDir = b.get('download_directory',dlPath)
    # self.dbPath = b['db_path']
    self.dbPath = b.get('db_path',
                        os.path.join(home,pmailDir,'pmail.db'))
    # print(self.dbPath)
    # self.logPath = b['log_path']
    self.logPath = b.get('log_path',
                         os.path.join(home,pmailDir,'pmail.log'))
    self.logLevel = b['log_level']
    self.updateFreq = b['update_frequency']
    self.port = b['port_number']
    
    am = y['appearance']['markers']
    ac = y['appearance']['colors']
    self.unread = am['unread']
    self.attachment = am['attachment']
    self.user = am['user']
    self.seperator = am['seperator']
    self.fg = ac['fg']
    self.bg = ac['bg']
    self.hiFg = ac['highlighted_fg']
    self.hiBg = ac['highlighted_bg']
    self.selFg = ac['selected_fg']
    self.selBg = ac['selected_bg']
    self.stFg = ac['statusline_fg']
    self.stBg = ac['statusline_bg']

    self.accounts = y['accounts']

    self.dbExists = False

    global config
    config = self

  def listAccounts(self):
    return self.accounts.keys()
  
  def listAccountIds(self):
    return [self.accounts[k]['id'] for k in self.listAccounts()]

  def getName(self, myemail):
    return self.accounts[myemail]['name']

  def fzfArgs(self, prompt):
    return self.picker.split(' ') + ['--prompt', prompt]

  def w3mArgs(self):
    return self.pager.split(' ')

  def vimArgs(self, draftId):
    args = re.split('(-c)', self.editor)
    file = 'f ' + os.path.join(self.tmpDir, draftId)
    args = list(filter(lambda x: not x == '', map(lambda x: x.strip(), args)))
    args.append(file)
    return args


# Parse config file
config = Config()

# If modifying these scopes, delete the file token.pickle.
SCOPES = ['https://www.googleapis.com/auth/gmail.readonly',
          'https://www.googleapis.com/auth/gmail.send',
          'https://www.googleapis.com/auth/gmail.modify',
          'https://www.googleapis.com/auth/gmail.settings.basic']


DB_PATH = 'sqlite:///' + os.path.join(WORKING_DIR ,config.dbPath)

HEADERS = ['From', 'Subject', 'To', 'Reply-To', 'In-Reply-To',
           'References', 'Message-ID', 'Content-Type']

logging.basicConfig(filename=os.path.join(WORKING_DIR, config.logPath),
                    level=logging.CRITICAL,
                    format='%(levelname)s::%(asctime)s::[%(module)s]: %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger(__name__)

engine = create_engine(DB_PATH)
Base = declarative_base(bind=engine)

session_factory = sessionmaker(bind=engine, 
    expire_on_commit=False)
Session = scoped_session(session_factory)

# <---

# ---> mkService


def mkService(account):
  '''
  Make a service object for use with the API.
  '''
  credentialsPath = config.accounts[account]['credentials']
  tokenPath = os.path.join(config.pickleDir, account.split('@')[0] + '.pickle')
  creds = None
  # The file *.pickle stores the user's access and refresh tokens, and is
  # created automatically when the authorization flow completes for the first
  # time.
  if os.path.exists(tokenPath):
    with open(tokenPath, 'rb') as token:
      creds = pickle.load(token)
  # If there are no (valid) credentials available, let the user log in.
  if not creds or not creds.valid:
    if creds and creds.expired and creds.refresh_token:
      creds.refresh(Request())
    else:
      flow = InstalledAppFlow.from_client_secrets_file(
          credentialsPath, SCOPES)
      creds = flow.run_local_server(port=8686)
    # Save the credentials for the next run
    with open(tokenPath, 'wb') as token:
      pickle.dump(creds, token)

  service = build('gmail', 'v1', credentials=creds, cache=MemoryCache())
  return service

# <---

# ---> Formating functions


def formatSize(num):
  '''
  Format a filesize in bytes into a human readable 
  one.
  '''
  for unit in ['B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
    if abs(num) < 1024.0:
      # return "%3.1f%s%s" % (num, unit, suffix)
      numString = "{:.1f}".format(num)
      lenNumString = len(numString)
      if lenNumString > 4:
        numString = numString[:4]
      elif lenNumString < 4:
        numString = numString + " " * (4 - lenNumString)
      return numString + unit
    num /= 1024.0
  # Unlikely to need this one!
  return "%.1f%s%s" % (num, 'Yi')


def unix2localTime(unix_timestamp):
  '''
  Convert timestamp to local time and format 
  like e.g. Jun 3
  '''
  utc_time = datetime.fromtimestamp(int(unix_timestamp)//1000, timezone.utc)
  local_time = utc_time.astimezone()
  return local_time.strftime("%b %d")


def unix2localTime2(unix_timestamp):
  '''
  Convert timestamp to local time and format like
  e.g. Mon, Jun 30, 2020 at 5.25 PM
  '''
  utc_time = datetime.fromtimestamp(int(unix_timestamp)//1000, timezone.utc)
  local_time = utc_time.astimezone()
  return local_time.strftime("%a, %b %d, %Y at %-I.%M %p")

# <---

# ---> Class defintions


class MemoryCache(Cache):
  '''
  This is needed so that the cache is working inside Thread 
  objects.
  '''
  _CACHE = {}

  def get(self, url):
    return MemoryCache._CACHE.get(url)

  def set(self, url, content):
    MemoryCache._CACHE[url] = content


class Attachments(Base):
  '''
  Store Attachments.
  '''
  __tablename__ = 'attachments'
  id = Column(Integer, primary_key=True)
  messageId = Column(String, ForeignKey('header_info.messageId'))
  partId = Column(String)
  filename = Column(String)
  contentType = Column(String)
  size = Column(Integer)

  def __init__(self, messageId, partId, filename, contentType, size):
    self.messageId = messageId
    self.partId = partId
    self.filename = filename
    self.contentType = contentType
    self.size = size

  def display(self):
    '''
    Display information about the attachment.
    '''
    contentType = self.contentType
    if len(contentType) > 20:
      contentType = contentType[:20]
    elif len(contentType) < 20:
      contentType = contentType + ' ' * (20 - len(contentType))
    display = '{} {} ({}) {}'.format(
        self.partId,
        contentType,
        formatSize(self.size),
        self.filename)
    return display


class AddressBook(Base):
  '''
  Store address in a seperate table so that it can be queried
  quickly rather than an expensive query on the MessageInfo
  table every time it is needed.
  '''
  __tablename__ = 'address_book'
  account = Column(String, ForeignKey('user_info.emailAddress'),
                   primary_key=True)
  address = Column(String, primary_key=True)

  def __init__(self, account, address):
    self.account = account
    self.address = address

  @classmethod
  def mk(cls, account, session, message):
    '''
    Method to scour for email address in the MessageInfo table
    and sort them into uniques and add to the AddressBook table
    '''
    q = session.query(cls).filter(cls.account == account)

    addresses = [a.address for a in q]
    newAddresses = set()

    sender = message.sender.lower().strip()
    if sender not in addresses:
      newAddresses.add(sender)
    try:
      all = filter(lambda x: not re.search(account, x),
                   (message.recipients).split(','))
      for a in all:
        address = a.lower().strip()
        if address not in addresses:
          newAddresses.add(address)
    except:
      pass

    # s.query(cls).filter(cls.account == account).delete()
    # s.commit()
    try:
      session.add_all([cls(account,a) for a in newAddresses])
      session.commit()
    except:
      logger.info('Something went wrong trying to add new addresses.')

  @classmethod
  def addressList(cls,account):
    session = Session()
    q = session.query(cls).filter(cls.account == account)
    Session.remove()
    for address in q:
      yield address.address


class UserInfo(Base):
  '''
  Store Profile information for each account.
  '''
  __tablename__ = 'user_info'
  emailAddress = Column(String, primary_key=True)
  totalMessages = Column(Integer)
  totalThreads = Column(Integer)
  historyId = Column(Integer)
  numOfUnreadMessages = Column(Integer)
  shouldIupdate = Column(Boolean)
  # token = Column(String)
  messages = relationship('MessageInfo', backref='user_info',
                          cascade='all, delete, delete-orphan')
  addressBook = relationship('AddressBook', backref='user_info',
                             cascade='all, delete, delete-orphan')

  def __init__(self, emailAddress, totalMessages, totalThreads, historyId,
               numOfUnreadMessages):
    self.emailAddress = emailAddress
    self.totalMessages = totalMessages
    self.totalThreads = totalThreads
    self.historyId = historyId
    self.numOfUnreadMessages = numOfUnreadMessages
    self.shouldIupdate = True


  @staticmethod
  def _numOfUnreadMessages(session, account):
    '''
    Get the number of unread messages in the INBOX.

    Args:
      account: The account to query.
    Returns: An integer.
    '''
    q1 = session.query(Labels.messageId).filter(Labels.label == 'UNREAD')
    q2 = session.query(Labels.messageId).filter(Labels.label == 'INBOX')
    q = session.query(MessageInfo).filter(
        MessageInfo.emailAddress == account,
        MessageInfo.messageId.in_(q1),
        MessageInfo.messageId.in_(q2))
    count = q.count()
    return count

  @classmethod
  def succesfulUpdate(cls, session, account):
    q = session.query(cls).get(account)
    q.shouldIupdate = False
    session.commit()

  @classmethod
  def update(cls, session, account, service, lastHistoryId):
    '''
    Method to update UserInfo.

    Args:
      account: emailAddress for which we are updating info.
      session: A DB session
      service: API service. Possibly None, in this case we only update
      numOfUnreadMessages and possibly the lastHistoryId.
      lastHistoryId: The history id from the last time the db was updated. 
      Possibly None, in which case do not update!
    '''
    numOfUnreadMessages = cls._numOfUnreadMessages(session, account)
    q = session.query(cls).get(account)
    if service:
      profile = service.users().getProfile(userId='me').execute()
      messagesTotal = profile['messagesTotal']
      threadsTotal = profile['threadsTotal']
      # historyId = profile['historyId']
      if q:
        q.messagesTotal = messagesTotal
        q.threadsTotal = threadsTotal
        q.historyId = lastHistoryId
        q.numOfUnreadMessages = numOfUnreadMessages
      else:
        userInfo = cls(account, messagesTotal, threadsTotal, lastHistoryId,
                   numOfUnreadMessages)
        session.add(userInfo)
    elif q:
      q.numOfUnreadMessages = numOfUnreadMessages
      if lastHistoryId:
        q.historyId = lastHistoryId
    session.commit()


class Labels(Base):
  '''
  Store Labels attached to messages.
  '''
  __tablename__ = 'labels'
  id = Column(Integer, primary_key=True)
  messageId = Column(String, ForeignKey('header_info.messageId'))
  label = Column(String)

  def __init__(self, messageId, label):
    self.messageId = messageId
    self.label = label

  @classmethod
  def addLabels(cls, session, labels):
    '''
    Add labels to messages.

    Args:
      session: A DB session.
      labels: A list of pairs (m,ls) where m is a message
      id and ls is a list of labels to add to the message.

    Returns: None.
    '''
    for (messageId, ls) in labels:
      for l in ls:
        label =cls(messageId, l)
        session.add(label)
        session.commit()

  @classmethod
  def removeLabels(cls, session, labels):
    '''
    Remove labels to messages.

    Args:
      session: A DB session.
      labels: A list of pairs (m,ls) where m is a message
      id and ls is a list of labels to add to the message.

    Returns: None.
    '''
    for (messageId, ls) in labels:
      session.query(cls).filter(cls.messageId == messageId,cls.label.in_(ls))\
          .delete(synchronize_session=False)
      session.commit()

# ---> MessageInfo class


class MessageInfo(Base):
  '''
  Big class for storing information about an email
  '''
  __tablename__ = 'header_info'
  messageId = Column(String, primary_key=True)
  emailAddress = Column(String, ForeignKey('user_info.emailAddress'))
  historyId = Column(String)
  time = Column(String, index=True)
  size = Column(Integer)
  snippet = Column(String)
  externalId = Column(String)
  subject = Column(String)
  sender = Column(String)
  replyTo = Column(String)
  inReplyTo = Column(String)
  references = Column(String)
  recipients = Column(String)
  contentType = Column(String)
  hasAttachments = Column(Boolean)
  labels = relationship('Labels', backref='header_info',
                        cascade="all, delete, delete-orphan",
                        lazy='subquery')
  attachments = relationship('Attachments', backref='header_info',
                             cascade="all, delete, delete-orphan")

  def __init__(self, messageId, emailAddress, historyId, time, size,
               snippet, externalId, subject, sender, replyTo, inReplyTo,
               references, recipients, contentType):
    self.messageId = messageId
    self.emailAddress = emailAddress
    self.historyId = historyId
    self.time = time
    self.size = size
    self.snippet = snippet
    self.externalId = externalId
    self.subject = subject
    self.sender = sender
    self.replyTo = replyTo
    self.inReplyTo = inReplyTo
    self.references = references
    self.recipients = recipients
    self.contentType = contentType

  def __eq__(self, other):
      if isinstance(other, self.__class__):
          return self.messageId == other.messageId
      else:
          return False

  def __ne__(self, other):
      return not self.__eq__(other)

  def display(self, senderWidth, scrWidth):
    '''
    For displaying information about a message in the main screen
    of the program.
    Args:
      senderWidth: How wide should the sender column be.
      scrWidth: How wide is the screen.

    Reurns:
      A formatted string.
    '''
    if 'UNREAD' in [l.label for l in self.labels]:
      marker = chr(config.unread) + ' '
    else:
      marker = '  '
    if self.existsAttachments():
      attachment =chr(config.attachment) + ' '
    else:
      attachment = '  '
    str = ' {} {} {} ({}) {} {}'.format(
        marker,
        unix2localTime(self.time),
        self.formatName(senderWidth),
        formatSize(self.size),
        attachment,
        self.subject,
        ' ' * scrWidth
    )[:scrWidth]
    return str

  def showLabels(self):
    return [l.label for l in self.labels]

  def timeForReply(self):
    '''
    Returns the formatted time for use in the extra line of
    information inserted at the top of replies.
    '''
    return unix2localTime2(self.time)

  def parseSender(self):
    '''
    Try to extract name and email address from the 'From' 
    field of the header.
    '''
    try:
      email = re.search('\<(.*?)\>', self.sender).group(1)
    except:
      email = self.sender
    name = re.sub(' \<(.*?)\>', '', self.sender)
    return email, name

  def formatName(self, senderWidth):
    name = self.parseSender()[1]
    name = re.sub('"', '', name)
    if len(name) < senderWidth:
      name = name + ' ' * (senderWidth - len(name))
    else:
      name = name[:senderWidth]
    return name

  def existsAttachments(self):
    '''
    Try to figure out if there are attachments from the content-type
    header. This can return false positives. 
    To deal with this if at some later stage (usually, when trying to view 
    attachments) it is found that really there were no attachments, the 
    hasAttachments attribute gets updated to False.
    '''
    if self.hasAttachments == False:
      return self.hasAttachments
    else:
      try:
        s = re.search('multipart/mixed', self.contentType)
      except:
        s = False
      if s == None:
        return False
      else:
        return True

  @classmethod
  def addMessage(cls, account, session, msg):
    '''
    Add a message to the db.
    Args:
      account: Account which owns the message.
      session: A db session.
      msg: The message to add.
    '''
    headers = dict(zip(HEADERS, [None for i in range(len(HEADERS))]))
    for h in msg['payload']['headers']:
      name = h['name']
      for n in HEADERS:
        if name.lower() == n.lower():
          headers[n] = h['value']

    try:
      labels = [Labels(msg['id'], label) for label in msg['labelIds']]
      session.add_all(labels)
    except KeyError:
      logger.debug('Could not add: ' + str(msg))
      logger.debug("Adding label ['UNCLASSIFIED']")
      session.add(Labels(msg['id'], 'UNCLASSIFIED'))
    if headers['From']:
      header =cls(
          msg['id'],
          account,
          msg['historyId'],
          msg['internalDate'],
          msg['sizeEstimate'],
          msg['snippet'],
          headers['Message-ID'],
          headers['Subject'],
          headers['From'],
          headers['Reply-To'],
          headers['In-Reply-To'],
          headers['References'],
          headers['To'],
          headers['Content-Type'])
      session.add(header)
      # logger.debug('Adding header to db')
      return header
    else:
      logger.debug(str(headers))
      logger.debug(str(msg))
      return None

  @classmethod
  def addMessages(cls, session, account, service, messageIds):
    '''
    Add many messages to the db.

    Args: 
      session: A db session.
      account: The accoun which owns the messages
      service: An API service.
      messageIds: List of message ids which we are going to add.
    '''
    def doesMessageAlreadyExist(messageId):
      q = session.query(cls).filter(cls.messageId == messageId)
      if q.first():
        return False
      else:
        return True

    def _updateDb(session, account, requestId, response, exception):
      '''
      Helper function for batch requests.
      
      Args:
        session: A DB session.
        account: The account which the messages belong to.
        requestId: Id of the request.
        response: Response of the request (If succesful, this is the message)
        exception: An exception if something went wrong.
      '''
      if exception is not None:
        logger.debug(exception)
        # Do something with the exception
        # See BatchHttpRequest docs
      else:
        # Parse headers.
        message = cls.addMessage(account, session, response)
        if message:
          AddressBook.mk(account, session, message)
        
    q = list(filter(doesMessageAlreadyExist, messageIds))
    i, l, batch = 0, len(q), service.new_batch_http_request()
    # logger.debug('Making batch request..')

    while i < l:
      for messageId in q[i:i+100]:
        batch.add(service.users().messages().get(
            userId='me',
            id=messageId,
            format='metadata',
            metadataHeaders=HEADERS
        ), callback=(lambda x, y, z: _updateDb(session, account, x, y, z)))
      batch.execute()
      batch = service.new_batch_http_request()
      i += 100
    session.commit()


  @classmethod
  def removeMessages(cls, session, messageIds):
    '''
    Delete messages (Not in use currently and 
    might need extra scopes to work.)
    '''
    session.query(cls).filter(cls.messageId in
                                messageIds).delete(synchronize_session=False)

# <---

# <---

# ---> Getting message Id's from Google


def listMessagesMatchingQuery(service, user_id, query=''):
  """List all Messages of the user's mailbox matching the query.

  Args:
    service: Authorized Gmail API service instance.
    user_id: User's email address. The special value "me"
    can be used to indicate the authenticated user.
    query: String used to filter messages returned.
    Eg.- 'from:user@some_domain.com' for Messages from a particular sender.

  Returns:
    List of Messages that match the criteria of the query. Note that the
    returned list contains Message IDs, you must use get with the
    appropriate ID to get the details of a Message.
  """
  try:
    response = service.users().messages().list(userId=user_id,
                                               q=query).execute()
    messages = []
    if 'messages' in response:
      messages.extend(response['messages'])

    while 'nextPageToken' in response:
      page_token = response['nextPageToken']
      response = service.users().messages().list(userId=user_id, q=query,
                                                 pageToken=page_token).execute()
      messages.extend(response['messages'])

    return messages
  except:
    print('An error occurred')

# <---

# ---> Db update functions



# <---

# ---> Find information

def countAttachments(message):
  c = 0
  if 'payload' in message:
    if 'parts' in message['payload']:
      for part in message['payload']['parts']:
        if part['filename']:
          c += 1
      return c
    else:
      return 0
  else:
    return 0


def myEmail(service):
  return (service.users().getProfile(userId='me').execute())['emailAddress']


# <---

# Only needed to create table structure
Base.metadata.create_all()

"""
vim:foldmethod=marker foldmarker=--->,<---
"""



#!/usr/bin/python

# ---> Imports

import pickle
import os.path
import os
import sys
import socket
import logging
import argparse

from apiclient import errors
from pmail.common import (mkService, Session, Labels, MessageInfo,
                    listMessagesMatchingQuery, logger,
                    UserInfo, AddressBook, config)
from googleapiclient.http import BatchHttpRequest
from threading import Thread, Lock, Event
from time import sleep

# <---

# ---> DB update and synchronisation


# ---> Pickling

def storeLastHistoryId(account, session, lastMessageId=None, lastHistoryId=None):
  '''
  Pickle the last historyId for use when updating the db.

  Args:
    account: Account whose history we are pickling.
    lastMessageId = The id of the last message saved in the db.
    lastHistoryId = The lastHistoryId.

  Returns:
    None
  '''
  # logger.info('Storing last historyId: {}'.format(lastHistoryId))
  # path = os.path.join(config.pickleDir, 'lastHistoryId.' + account + '.pickle')
  if lastMessageId:
    lastHistoryId = session.query(MessageInfo).get(lastMessageId).historyId
    q = session.query(UserInfo).get(account)
    q.historyId = lastHistoryId
    session.commit()
    # with open(path, 'wb') as f:
    #   pickle.dump(q.historyId, f)
  elif lastHistoryId:
    q = session.query(UserInfo).get(account)
    q.historyId = lastHistoryId
    session.commit()
    # with open(path, 'wb') as f:
    #   pickle.dump(lastHistoryId, f)


def getLastHistoryId(account, session):
  '''
  Get the last history for account from the pickle if it exists.
  '''
  # path = os.path.join(config.pickleDir, 'lastHistoryId.' + account + '.pickle')
  # if os.path.exists(path):
  #   with open(path, 'rb') as f:
  #     return pickle.load(f)
  return session.query(UserInfo).get(account).historyId


# <---

# ---> List changes since last update


def ListHistory(service, user_id, start_history_id='1'):
  """
  List History of all changes to the user's mailbox.

 Args:
    service: Authorized Gmail API service instance.
    user_id: User's email address. The special value "me"
    can be used to indicate the authenticated user.
    start_history_id: Only return Histories at or after start_history_id.

  Returns:
    A list of mailbox changes that occurred after the start_history_id.
  """
  try:
    history = (service.users().history().list(userId=user_id,
                                              startHistoryId=start_history_id)
               .execute())
    changes = history['history'] if 'history' in history else []
    while 'nextPageToken' in history:
      page_token = history['nextPageToken']
      history = (service.users().history().list(userId=user_id,
                                                startHistoryId=start_history_id,
                                                pageToken=page_token).execute())
      changes.extend(history['history'])

    return changes
  except errors.Error as e:
    logger.debug(e)

# <---


def updateDb(account, session, service, newMessagesArrived, lastHistoryId=None):
  '''
  Update the Database with messages since lastHistoryId.

  Args:
    account: The account which the messages come from.
    session: DB session.
    service: The google API sevice object.
    lastHistoryId: The historyId of the last update, if there was one.

  Returns:
    historyId corresponding to this update.
  '''
  # 100 is maximum size of batch!
  if lastHistoryId is None:
    messageIds = listMessagesMatchingQuery(service(account),
                                           'me', query='newer_than:' + config.syncFrom)
    MessageInfo.addMessages(session, account,
                            service(account), [m['id'] for m in messageIds])
    lastMessageId = messageIds[0]['id']
    # logger.info('lastMessageId: {}'.format(lastMessageId))

  else:
    changes = ListHistory(service(account), 'me',
                          getLastHistoryId(account, session))
    messagesAdded, messagesDeleted = [], []
    labelsAdded, labelsRemoved = [], []

    for change in changes:
      if 'messagesAdded' in change:
        messagesAdded += [c['message']['id'] for c in
                          change['messagesAdded']]
      if 'messagesDeleted' in change:
        messagesDeleted += [c['message']['id'] for c in
                            change['messagesDeleted']]
      if 'labelsAdded' in change:
        labelsAdded += [(c['message']['id'], c['labelIds']) for c in
                        change['labelsAdded']]
      if 'labelsRemoved' in change:
        labelsRemoved += [(c['message']['id'], c['labelIds']) for c in
                          change['labelsRemoved']]

    MessageInfo.addMessages(session, account, service(account), messagesAdded)
    if len(messagesAdded) > 0:
      # Set newMessagesArrived to be true, this will cause the Cache
      # (savedQuery) to be updated.
      logger.info('There are new messages!')
      newMessagesArrived.set()

    MessageInfo.removeMessages(session, messagesDeleted)
    Labels.addLabels(session, labelsAdded)
    Labels.removeLabels(session, labelsRemoved)
    # try:
    if len(changes) > 0:
      lastHistoryId = str(max([int(change['id']) for change in changes]))
      # logger.info('Last history id: {}'.format(lastHistoryId))
    else:
    # except Exception as e:
      # logger.debug(e)
      lastHistoryId = None
    lastMessageId = None
  # storeLastHistoryId(account, session, lastMessageId=lastMessageId,
  #                    lastHistoryId=lastHistoryId)
  if lastMessageId:
    lastHistoryId = session.query(MessageInfo).get(lastMessageId).historyId
    # logger.info('lastHistoryId: {}'.format(lastHistoryId))
  return lastHistoryId

# <---

# ---> Server Functions


def getMessages(s, Q, newMessagesArrived, account, query, position,
                height, excludedLabels=[],
                includedLabels=[], count=False, afterAction=None):
  '''
  Get a list of messages to display.

  Args:
    account: The currently selected account.
    query: list of messageIds.
    position: The position of the currently highlighted message.
    height: The height of the stdscr.
    excludedLabels: Any labels to exclude,
    includedLabels: Any labels to include.
    count: if True then only return the count.

  Returns:
    Either a list of MessageInfo objects or
    an integer (depending on truthiness of count).
  '''
  if newMessagesArrived.is_set():
    logger.info('New messages arrived, refreshing the cache.')
    q = Q.getQuery(s, account, query, includedLabels,
                   excludedLabels, refresh=True)
    newMessagesArrived.clear()
  elif afterAction == None:
    q = Q.getQuery(s, account, query, includedLabels, excludedLabels)
  elif afterAction['action'] in ['DELETE', 'TRASH']:
    logger.info('Removing message: {} from cache.'
                .format(afterAction['messageIds']))
    q = Q.removeMessages(afterAction['messageIds'])
  elif afterAction['action'] in ['MARK_AS_READ']:
    # logger.info('removing msg from cache')
    logger.info('Refreshing the cache, after reading.')
    # q = Q.markAsRead(afterAction['messageIds'])
    q = Q.getQuery(s, account, query, includedLabels,
                   excludedLabels, refresh=True)
  if count == False:
    # return [h for h in q.slice(position, position + height - 2)]
    return q[position:position + height - 2]
  elif count == True:
    # return q.count()
    return len(q)


class SaveQuery():

  def __init__(self):
    self.account = None
    self.query = None
    self.includedLabels = None
    self.excludedLabels = None
    self.savedQuery = None

  def getQuery(self, s, account, query, includedLabels, excludedLabels, refresh=False):
    # Also deletes, new mails, mark as read, etc..
    # logger.info("read :{}.".format(read))
    if (self.account == account and
        str(self.query) == str(query) and
        self.includedLabels == includedLabels and
            self.excludedLabels == excludedLabels) and refresh == False:
      # logger.info('Using saved query.')
      return self.savedQuery
    else:
      excludeQuery = s.query(Labels.messageId).filter(
          Labels.label.in_(excludedLabels))
      includeQuery = s.query(Labels.messageId).filter(
          Labels.label.in_(includedLabels))
      # logger.info('Made two queries.')
      q = s.query(MessageInfo)\
          .filter(
              MessageInfo.messageId.in_(query),
              MessageInfo.emailAddress == account,
              ~MessageInfo.messageId.in_(excludeQuery),
              MessageInfo.messageId.in_(includeQuery))\
          .order_by(MessageInfo.time.desc())
      # logger.info('Successfully(?) refreshed the cache.')
      self.account = account
      self.query = query
      self.includedLabels = includedLabels
      self.excludedLabels = excludedLabels
      self.savedQuery = list(q)
      # logger.info('Generating new query.')
      return self.savedQuery

  def removeMessages(self, messageIds):
    self.savedQuery = [m for m in self.savedQuery if m.messageId not in
                       messageIds]
    return self.savedQuery

  # This didn't work - probably remove it later
  def markAsRead(self, messageIds):
    logger.info('Marking as read.')
    newSavedQuery = []
    for m in self.savedQuery:
      logger.info('trying')
      if m.messageId in messageIds:
        logger.info('Found message with labels: {}'
                    .format(m.labels))
        m.labels = [l.label for l in m.labels if l.label != 'UNREAD']
        logger.info('setting read to True on msg: {}'.format(m.messageId))
      newSavedQuery.append(m)
    self.savedQuery = newSavedQuery
    return self.savedQuery


'''
def getNextMessage(s, account, query, lastTime, excludedLabels=[],
                includedLabels=[]):
  Get a list of messages to display.

  Args:
    account: The currently selected account.
    query: list of messageIds.
    position: The position of the currently highlighted message.
    height: The height of the stdscr.
    excludedLabels: Any labels to exclude,
    includedLabels: Any labels to include.
    count: if True then only return the count.

  Returns:
    Either a list of MessageInfo objects or
    an integer (depending on truthiness of count).
  excludeQuery = s.query(Labels.messageId).filter(
      Labels.label.in_(excludedLabels))
  includeQuery = s.query(Labels.messageId).filter(
      Labels.label.in_(includedLabels))
  q = s.query(MessageInfo)\
      .filter(
          MessageInfo.messageId.in_(query),
          MessageInfo.emailAddress == account,
          ~MessageInfo.messageId.in_(excludeQuery),
          MessageInfo.messageId.in_(includeQuery))\
      .filter(MessageInfo.time < lastTime)\
      .order_by(MessageInfo.time.desc())
  return q.first() 

'''

# <---

# ---> Main


def pmailServer(lock, newMessagesArrived, Q):
    # get the hostname
  host = socket.gethostname()
  port = config.port
  bufferSize = 1024

  portFree = True
  while portFree == True:
    try:
      sock = socket.socket()
      sock.bind((host, port))
      portFree = False
      logger.info('Listening on port {}.'
                  .format(config.port))
    except OSError as e:
      logger.info(e)
      logger.warning('Port {} busy. Checking port again in 30s.'
                  .format(config.port))
      sleep(30)

  s = Session()

  # configure how many client the server can listen simultaneously
  # only accept connections from client?
  sock.listen(1)
  while 1:
    conn, address = sock.accept()
    print("Connection from: {}.".format(str(address)))
    sizeOfIncoming = int.from_bytes(conn.recv(4), 'big')
    print(sizeOfIncoming)
    incoming = b''

    while len(incoming) < sizeOfIncoming:
      incoming += conn.recv(bufferSize)

    unpickledIncoming = pickle.loads(incoming, encoding='bytes')

    # Do something and return a response.
    action = unpickledIncoming['action']
    with lock:
      if action == 'CHECK_FOR_NEW_MESSAGES':
        if newMessagesArrived.is_set():
          response = 'newMessagesArrived'
          # newMessagesArrived.clear()
        else:
          response = 'noNewMessages'
      elif action == 'GET_MESSAGES':
        # Get messages.
        incomingQ = unpickledIncoming['query']
        query = s.query(
            MessageInfo.messageId) if incomingQ is None else incomingQ
        # logger.info('afterAction: {}'.format(unpickledIncoming['afterAction']))
        response = getMessages(s,
                               Q,
                               newMessagesArrived,
                               unpickledIncoming['account'],
                               query,
                               unpickledIncoming['position'],
                               unpickledIncoming['height'],
                               unpickledIncoming['excludedLabels'],
                               unpickledIncoming['includedLabels'],
                               unpickledIncoming['count'],
                               unpickledIncoming['afterAction'])
      elif action == 'ADD_MESSAGES':
        # Add messages.
        account = unpickledIncoming['account']
        messageIds = unpickledIncoming['messageIds']
        MessageInfo.addMessages(s, account, mkService(account), messageIds)
        response = None
      elif action == 'GET_QUERY':
        # get a query.
        messageIds = unpickledIncoming['messageIds']
        cls = unpickledIncoming['class']
        response = [e for e in s.query(cls)
                    .filter(cls.messageId.in_(messageIds))]
      elif action == 'REMOVE_LABELS':
        # logger.info('Removing labels...')
        # Remove labels.
        # logger.info('Recieved instructions from client'
        #             + '{}'.format(unpickledIncoming))
        labels = unpickledIncoming['labels']
        Labels.removeLabels(s, labels)
        try:
          account = unpickledIncoming['account']
          incomingQ = unpickledIncoming['query']
          query = s.query(
              MessageInfo.messageId) if incomingQ is None else incomingQ
          # query = unpickledIncoming['query']
          includedLabels = unpickledIncoming['includedLabels']
          excludedLabels = unpickledIncoming['excludedLabels']
          # logger.info('Refreshing saved query.')
          Q.getQuery(s, account, query, includedLabels,
                     excludedLabels, refresh=True)
        except Exception as e:
          logger.warning(e)
        response = None
        # response = Q.markAsRead([l[0] for l in labels])
      elif action == 'ADD_LABELS':
        # Add labels.
        # logger.info('Recieved instructions from client'
        #             + '{}'.format(unpickledIncoming))
        labels = unpickledIncoming['labels']
        Labels.addLabels(s, labels)
        response = None
      elif action == 'REMOVE_FALSE_ATTACMENTS':
        # Remove false attachments.
        messageId = unpickledIncoming['messageId']
        s.query(MessageInfo).filter(MessageInfo.messageId == messageId)\
            .update({MessageInfo.hasAttachments: False},
                    synchronize_session='evaluate')

    pickledResponse = pickle.dumps(response)
    sizeOfPickle = len(pickledResponse)
    numOfChunks = sizeOfPickle//bufferSize + 1
    conn.send(sizeOfPickle.to_bytes(4, 'big'))

    for i in range(numOfChunks):
      conn.send(pickledResponse[i*bufferSize: (i+1)*bufferSize])

    # if action == 'REMOVE_LABELS':
    #   for account in config.listAccounts():
    #     UserInfo.update(s, account, None, None)
    s.commit()
    s.close()
    print('closing connection..')
    conn.close()


def syncDb(lock, newMessagesArrived):
  s = Session()
  while 1:
    try:
      with lock:
        logger.info('Acquired lock, making session.')
        for account in config.listAccounts():
          shouldIupdate = s.query(UserInfo).get(account)
          # logger.info('shouldIupdate: {}'.format(shouldIupdate))
          # logger.info('shouldIupdate.shouldIupdate: {}'
          #              .format(shouldIupdate.shouldIupdate))

          if shouldIupdate and shouldIupdate.shouldIupdate == False:
            # logger.info('Performing partial update for {}.'.format(account))
            lastHistoryId = updateDb(account, s, mkService,
                                     newMessagesArrived, getLastHistoryId(account, s))
            UserInfo.update(s, account, None, lastHistoryId)
            logger.info('Successful partial update for {}.'.format(account))

          elif shouldIupdate is None or shouldIupdate.shouldIupdate == True:
            # logger.info('Performing full update for {}.'.format(account))
            lastHistoryId = updateDb(account, s, mkService, newMessagesArrived)
            UserInfo.update(s, account, mkService(account), lastHistoryId)
            UserInfo.succesfulUpdate(s, account)
            logger.info('Successful full update for {}.'.format(account))
            with open(os.path.join(config.pickleDir,
                                   'synced.pickle'), 'wb') as f:
              pickle.dump(False, f)
      s.close()
    except Exception as e:
      logger.warning(e)
      logger.info('Unable to update DB, trying again soon.')
    except KeyboardInterrupt as k:
      print("Bye!")
      logger.info(k)
      sys.exit()
    sleep(config.updateFreq)


# if __name__ == '__main__':
def start():
  logger.setLevel(logging.DEBUG)
  # Uncomment this for sql logs
  # logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

  # parser = argparse.ArgumentParser()
  # parser.add_argument('-n', action='store')
  # args = parser.parse_args()
  Q = SaveQuery()

  # if args.n == None:
  lock = Lock()
  newMessagesArrived = Event()
  t1 = Thread(target=pmailServer, 
              args=(lock, newMessagesArrived, Q),
              daemon=True)
  t2 = Thread(target=syncDb, args=(lock, newMessagesArrived,))
  t1.start()
  t2.start()

def checkForNewMessages(id):
  for k in config.listAccounts():
    if config.accounts[k]['id'] == id:
      account = k
  s = Session()
  q = s.query(UserInfo.numOfUnreadMessages)\
      .filter(UserInfo.emailAddress == account)
  Session.remove()
  return(q[0][0])

# <---

"""
vim:foldmethod=marker foldmarker=--->,<---
"""



import argparse
import textwrap
import os, sys
from pmail.common import config, WORKING_DIR
from shutil import copyfile
import pmail.client
import pmail.server

if __name__ == '__main__':
  parser = argparse.ArgumentParser(prog='pmail',
                                   description='simple tui client for gmail',
                                   formatter_class=argparse.RawDescriptionHelpFormatter, epilog=textwrap.dedent('''\
pmail must be run with EITHER of the flags -m or -n to do something useful
if pmail is run with BOTH of -m and -n flags then it will exit with a warning
'''))

  parser.add_argument('-v', '--version', action='version',
                    version='%(prog)s ' + pmail.__version__)

  configDir = os.path.join(os.environ['HOME'],'.config','pmail')
  configPath = os.path.join(configDir, 'config.yaml')

  parser.add_argument('--mk-config',
                      help='Install an example file to: {}'
                      .format(configPath),
                      action='store_true')
  
  # parser.add_argument('-c',
  #                     # '--config',
  #                     metavar='PATH',
  #                     help='PATH=path/to/config.yaml',
  #                     # default='config.yaml',
  #                     action='store')

  parser.add_argument('-n',
                      metavar='ACCOUNT_ID',
                      choices=config.listAccountIds(),
                      help='return the number of unread messages for ACCOUNT_ID',
                      action='store')


  parser.add_argument('-m',
                      metavar='MODE',
                      choices=['server', 'client'],
                      help='run pmail in MODE=[server|client] mode',
                      action='store')

  args = parser.parse_args()

  if args.mk_config:
    if not os.path.exists(configPath):
      if not os.path.exists(configDir):
        os.makedirs(configDir)
      configSrc = os.path.join(WORKING_DIR,'config.yaml')
      copyfile(configSrc, configPath)
      print('Config successfully installed at: {}.'
            .format(configPath))
    else:
      print('Config file already exists at: {}.'
            .format(configPath))
    sys.exit()

  if args.n == None and args.m == None:
    parser.print_help()
  elif args.n and args.m:
    msg = 'pmail can only be run with ONE of -n or -m'
    raise argparse.ArgumentTypeError(msg)
  elif args.n:
    # print(account)
    pmail.server.checkForNewMessages(args.n)
    # print(args.n)
  elif args.m == 'client':
    pmail.client.start()
  elif args.m == 'server':
    pmail.server.start()
